/***********************************************************************

    Copyright (c) Johann Schoonees 2004 - 2015

    This file is part of Camwire, a generic camera interface.

    Camwire is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2.1 of the
    License, or (at your option) any later version.

    Camwire is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Camwire; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA

   
    Title: Camwire bus module

    Description:
    This Camwire bus module is about finding all visible cameras and
    providing handles to them.  The handles should be all a user need
    know about for complete access to all camera functions.  Functions
    to control individual cameras through their handles are defined in
    the Camwire main module.

***********************************************************************/


#include <stdlib.h>  /* calloc, malloc, free */

#include "dc1394/dc1394.h"

#include "camwire/camwire_handle.h"  /* Camwire_handle */
#include "camwire/camwirebus.h"


/* Local status: */
static int num_cams = 0;
static Camwire_handle * handle_array = 0;
static dc1394_t * dc1394_lib = 0;


/*
  ----------------------------------------------------------------------
  See camwirebus.h for documentation on this function.
*/
/* This can be implemented as the constructor of a Camwire_bus class. */
/* Camwire_handle is defined in camwirebus.h.*/

Camwire_handle * camwire_bus_create(int *num_handles)
{
    int h;
    int camcount;
    dc1394camera_list_t *list = 0;
    dc1394error_t err;
    Camera_handle camera;

    /* Check whether the bus has not already been created: */
    if (camwire_bus_exists())
    {
	*num_handles = num_cams;
	return handle_array;
    }

    /* Set default (error) values: */
    *num_handles = -1;
    num_cams = 0;

    /* Initialize the dc1394 library: */
    dc1394_lib = dc1394_new();
    if (!dc1394_lib)  return 0;

    /* Make a list of the visible cameras: */
    err = dc1394_camera_enumerate(dc1394_lib, &list);
    if (err != DC1394_SUCCESS)
    {
	dc1394_free(dc1394_lib);
	dc1394_lib = 0;
	return 0; 	/* Error accessing a camera.*/
    }

    camcount = list->num;
    if (camcount == 0) 	/* Nothing wrong, just found no cameras.*/
    {
	dc1394_camera_free_list(list);
	dc1394_free(dc1394_lib);
	dc1394_lib = 0;
	*num_handles = 0;
	return 0;
    }

    /* Allocate memory for camwire handles: */
    handle_array = (Camwire_handle *) calloc((size_t) camcount, sizeof(Camwire_handle));
    if (handle_array == 0)
    {
	dc1394_camera_free_list(list);
	dc1394_free(dc1394_lib);
	dc1394_lib = 0;
	return 0; 	/* Allocation failure.*/
    }

    /* Fill in the camera handle list: */
    for (h = 0; h<camcount; h++)
    {
	camera = dc1394_camera_new(dc1394_lib, list->ids[h].guid);
	if (camera)
	{
	    handle_array[num_cams] =
		(Camwire_bus_handle *) malloc(sizeof(Camwire_bus_handle));
	    if (!handle_array[num_cams])
	    { 	/* Allocation failure.*/
		dc1394_camera_free_list(list);
		camwire_bus_destroy();
		return 0;
	    }
	    handle_array[num_cams]->camera = camera;
	    handle_array[num_cams]->userdata = 0;
	    num_cams++;
	}
    }
    
    dc1394_camera_free_list(list);
    if (num_cams == 0)  camwire_bus_destroy();
    
    /* Returns: */
    *num_handles = num_cams;
    return handle_array;
} /* camwire_bus_create() */

/*
  ----------------------------------------------------------------------
  See camwirebus.h for documentation on this function.
*/
int camwire_bus_exists(void)
{
    return (dc1394_lib && handle_array && num_cams >= 0);
}

/*
  ----------------------------------------------------------------------
  See camwirebus.h for documentation on this function.
*/
/* This can be implemented as the destructor of a Camwire_bus class. */

void camwire_bus_destroy(void)
{
    int h;
    Camera_handle last_camera = 0;

    for (h=0; h<num_cams; h++)
    {
	/* Destroy camera handles, skipping duplicates: */
	if ((h == 0) || (handle_array[h]->camera != last_camera)) 
	{
	    last_camera = handle_array[h]->camera;
	    dc1394_camera_free(handle_array[h]->camera);
	}
	free(handle_array[h]);
    }
    free(handle_array);
    handle_array = 0;
    num_cams = 0;
    if (dc1394_lib)  dc1394_free(dc1394_lib);
    dc1394_lib = 0;
} /* camwire_bus_destroy() */

/*
  ----------------------------------------------------------------------
  See camwirebus.h for documentation on this function.
*/
void camwire_bus_reset(void)
{
    int bus_existed;
    Camwire_handle *handle_list;
    int num_cameras;
    int h;
    Camera_handle last_camera = 0;

    bus_existed = camwire_bus_exists();
    handle_list = camwire_bus_create(&num_cameras);
    if (camwire_bus_exists())
    {
	for (h = 0; h<num_cameras; h++)
	{
	    if ((h == 0) || (handle_list[h]->camera != last_camera)) 
	    {
		last_camera = handle_list[h]->camera;
		dc1394_reset_bus(handle_list[h]->camera);
	    }
	}
    }
    if (!bus_existed)  camwire_bus_destroy();
} /* camwire_bus_reset */

/*
  ----------------------------------------------------------------------
   Returns the dc1394camera_t camera handle for the given camwire
   handle.  Needed by many dc1394 functions in Camwire.
*/
Camera_handle camwire_handle_get_camera(const Camwire_handle c_handle)
{
    return c_handle->camera;
}

/*
  ----------------------------------------------------------------------
   Returns a pointer to the user data structure for the given camwire
   handle.  Needed for internal status maintenance in Camwire.
*/
User_handle camwire_bus_get_userdata(
    const Camwire_handle c_handle)
{
    return c_handle->userdata;
}

/*
  ----------------------------------------------------------------------
   Registers a pointer to a user data structure for the given camwire
   handle.  Returns 1 on success and 0 if the handle does not exist or
   if the pointer is already in use.  Needed for internal status
   maintenance in Camwire.
*/
int camwire_bus_set_userdata(const Camwire_handle c_handle,
			User_handle user_data)
{
    if (!c_handle)  return 0;
    if (user_data && c_handle->userdata)  return 0;
    c_handle->userdata = user_data;
    return 1;
}
