/***********************************************************************

    Copyright (c) Industrial Research Limited 2004

    This file is part of Camwire, a generic camera interface.

    Camwire is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2.1 of the
    License, or (at your option) any later version.

    Camwire is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Camwire; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA


    Title: Simple SDL display server

    Description:
    This module uses SDL to display images.

***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <SDL/SDL.h>

#include "display-codes.h"
#include "display-sdl.h"


#define MAX_IMAGE_FILE_HEADER	50


/* The display info points to this struct: */
typedef struct display_info_data
{
    SDL_Surface *screen;
    SDL_Surface *frame;
    int component_size;
    unsigned long maxval;
    void *aux_buffer;
}
Display_info_data;


/* Local functions: */
static int pixel_coding_supported(const Camwire_pixel coding);
static int equivalent_8bit(const int cam_depth);
static SDL_Surface * create_display_surface(const int width,
					  const int height,
					  const int depth);
static SDL_Surface * create_frame_surface(const int width,
					  const int height,
					  const int depth);
static void set_linear_palette(SDL_Surface *surface);

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
Display_info display_sdl_create(const int width,
				const int height,
				const Camwire_pixel coding,
				const unsigned long maxval)
{
    Display_info info;
    int camera_depth, display_depth;

    /* First check supported pixel codings and get depth:*/
    if (!pixel_coding_supported(coding))
    {
	fprintf(stderr, "Requested pixel coding is not supported in "
		"display_sdl_create()\n");
	return(NULL);
    }
    camwire_pixel_depth(coding, &camera_depth);
    display_depth = equivalent_8bit(camera_depth);  /* SDL only deals in
						       8-bit images.*/
    
    /* Create our handle: */
    info = (Display_info_data *) malloc(sizeof(Display_info_data));
    if (info == NULL)  return(NULL);

    /* Initialize SDL: */
    if (SDL_Init(SDL_INIT_VIDEO) == -1)
    {
	fprintf(stderr,
		"SDL_Init() failed in display_sdl_create():\n"
		"%s\n", SDL_GetError());
	free(info);
	return(NULL);
    }

    /* Set up the screen display and video mode:*/
    info->screen = create_display_surface(width, height, display_depth);
    if (info->screen == NULL)
    {
	SDL_Quit();
	free(info);
	return(NULL);
    }

    /* Create an empty frame surface.  display_sdl_frame() will assign
       the frame buffer pointer to it: */
    info->frame = create_frame_surface(width, height, display_depth);
    if(info->frame == NULL)
    {
	SDL_Quit(); 	/* Also frees info->screen.*/
	free(info);
	return(NULL);
    }

    /* Remember the camera pixel component size and max value for when
       we display 16-bit images: */
    info->component_size = camera_depth/display_depth;
    info->maxval = maxval;

    /* Allocate an auxilliary buffer for displaying 16-bit images: */
    if (info->component_size != 1)
    {
	info->aux_buffer = malloc((size_t)width*height*display_depth/8);
	if (info->aux_buffer == NULL)
	{
	    SDL_Quit();
	    free(info);
	    return(NULL);
	}
    }
    else
    {
	info->aux_buffer = NULL;
    }

    return(info);
}

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
void display_sdl_destroy(Display_info info)
{
    free(info->aux_buffer);
    SDL_FreeSurface(info->frame);
    SDL_Quit(); 	/* Also frees info->screen.*/
    free(info);
}

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_sdl_resize(Display_info info, const int width,
		       const int height)
{
    int old_width, old_height, old_depth;
    int new_mode;

    /* Remember some current parameters:*/
    old_width = info->frame->w;
    old_height = info->frame->h;
    old_depth = info->frame->format->BitsPerPixel;

    /* New video mode: */
    new_mode = (width != old_width || height != old_height);
    if (new_mode)
    {
	info->screen = create_display_surface(width, height, old_depth);
	if (info->screen == NULL)
	{
	    info->screen = create_display_surface(old_width, old_height,
						  old_depth);
	    return(DISPLAY_FAILURE);
	}

	/* New frame surface: */
	SDL_FreeSurface(info->frame);
	info->frame = create_frame_surface(width, height, old_depth);
	if (info->frame == NULL)
	{  /* Bother.  Undo as best we can.*/
	    info->screen = create_display_surface(old_width, old_height,
						  old_depth);
	    info->frame =
		create_frame_surface(old_width, old_height, old_depth);
	    return(DISPLAY_FAILURE);
	}

	/* Take care of auxilliary buffer: */
	if (info->component_size != 1)
	{
	    info->aux_buffer =
		realloc(info->aux_buffer, (size_t)width*height*old_depth/8);
	    if (info->aux_buffer == NULL)
	    {  /* Bother.  Undo as best we can.*/
		info->aux_buffer = malloc((size_t)old_width*old_height*old_depth/8);
		info->screen = create_display_surface(old_width, old_height,
						      old_depth);
		SDL_FreeSurface(info->frame);
		info->frame = create_frame_surface(old_width, old_height, old_depth);
		return(DISPLAY_FAILURE);
	    }
	}
    }

    return(DISPLAY_SUCCESS);
}

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_sdl_coding(Display_info info, const Camwire_pixel coding)
{
    int old_width, old_height, old_depth, old_size;
    int camera_depth, display_depth;
    int new_mode;

    if (!pixel_coding_supported(coding))  return(DISPLAY_FAILURE);

    /* Remember some current parameters:*/
    old_width = info->frame->w;
    old_height = info->frame->h;
    old_depth = info->frame->format->BitsPerPixel;
    old_size = info->component_size;

    /* New frame surface if depth really has changed: */
    camwire_pixel_depth(coding, &camera_depth);
    display_depth = equivalent_8bit(camera_depth);
    info->component_size = camera_depth/display_depth;
    new_mode = (display_depth != old_depth);
    if (new_mode)
    {
	/* New video mode: */
	info->screen = create_display_surface(old_width, old_height, display_depth);
	if (info->screen == NULL)
	{
	    /* Undo: */
	    info->screen = create_display_surface(old_width, old_height, old_depth);
	    info->component_size = old_size;
	    return(DISPLAY_FAILURE);
	}

	/* New frame surface: */
	SDL_FreeSurface(info->frame);
	info->frame = create_frame_surface(old_width, old_height, display_depth);
	if (info->frame == NULL)
	{
	    /* Undo: */
	    info->screen = create_display_surface(old_width, old_height, old_depth);
	    info->frame = create_frame_surface(old_width, old_height, old_depth);
	    info->component_size = old_size;
	    return(DISPLAY_FAILURE);
	}
    }
    
    /* Take care of auxilliary buffer: */
    if (old_size != 1 && info->component_size == 1)
    {
	free(info->aux_buffer);
	info->aux_buffer = NULL;
    }
    else if (old_size == 1 && info->component_size != 1)
    {
	info->aux_buffer = malloc((size_t)old_width*old_height*display_depth/8);
    }
    else if (old_size != 1 && info->component_size != 1 && old_depth != display_depth)
    {
	info->aux_buffer =
	    realloc(info->aux_buffer, (size_t)old_width*old_height*display_depth/8);
    }

    if (info->component_size != 1 && info->aux_buffer == NULL)
    {
	/* Undo as best we can: */
	info->aux_buffer = malloc((size_t)old_width*old_height*old_depth/8);
	if (new_mode)
	{
	    info->screen = create_display_surface(old_width, old_height, old_depth);
	    SDL_FreeSurface(info->frame);
	    info->frame = create_frame_surface(old_width, old_height, old_depth);
	}
	return(DISPLAY_FAILURE);
    }
    
    return(DISPLAY_SUCCESS);
}

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_sdl_maxval(Display_info info, const unsigned long maxval)
{
    if (maxval > 0xffff)  return(DISPLAY_FAILURE);
    info->maxval = maxval;
    return(DISPLAY_SUCCESS);
}

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_sdl_frame(Display_info info, const void *buffer)
{
    SDL_Event event;
    size_t num_components;
    unsigned long maxval;
    int shift;
    Uint16 *src;
    Uint8 *dest, *dest_beg, *dest_end;

    /* Check for user quit event: */
    while (SDL_PollEvent(&event))
    {
	if (event.type == SDL_QUIT)
	{  /* Display window has been closed.*/
	    return(DISPLAY_QUIT_REQ);
	}
	else if (event.type == SDL_KEYDOWN)
	{
	    if (event.key.keysym.sym == SDLK_q)
	    {  /* User typed `Q'.*/
		return(DISPLAY_QUIT_REQ);
	    }
	}
    }

    /* Convert 16-bit to 8-bit components: */
    if (info->component_size != 1)
    {
	if (info->aux_buffer == NULL)
	{
	    fprintf(stderr,
		    "display_sdl_frame(): Internal error: info->aux_buffer is NULL.");
	}
	num_components =
	    (size_t)info->frame->w*info->frame->h*info->frame->format->BitsPerPixel/8;
	maxval  = info->maxval;
	shift = 0;
	while (maxval != 0)
	{
	    maxval >>= 1;
	    ++shift;
	}
	if (shift < 8 || shift > 16)  shift = 16;
	shift -= 8;
	src = (Uint16 *) buffer;
	dest_beg = (Uint8 *) info->aux_buffer;
	dest_end = dest_beg + num_components;
	for (dest=dest_beg; dest != dest_end; ++dest)  *dest = ntohs(*src++) >> shift;
	info->frame->pixels = info->aux_buffer;
    }
    else
    {
	info->frame->pixels = (void *) buffer;
    }

    /* Display it:*/
    SDL_BlitSurface(info->frame, NULL, info->screen, NULL);
    SDL_Flip(info->screen);
    /* Note: X11 does not support hardware acceleration.  All SDL_Flip()
       does is to update the screen. */
    
    return(DISPLAY_SUCCESS);
}

/*
  ----------------------------------------------------------------------
  Return 1 if pixel coding is supported, otherwise 0.
*/
static int pixel_coding_supported(const Camwire_pixel coding)
{
    switch (coding)
    {
	case CAMWIRE_PIXEL_MONO8:
	case CAMWIRE_PIXEL_RGB8:
	case CAMWIRE_PIXEL_MONO16:
	case CAMWIRE_PIXEL_RGB16:
	case CAMWIRE_PIXEL_MONO16S:
	case CAMWIRE_PIXEL_RGB16S:
	case CAMWIRE_PIXEL_RAW8:
	case CAMWIRE_PIXEL_RAW16:
	    return(1);
	case CAMWIRE_PIXEL_YUV411:
	case CAMWIRE_PIXEL_YUV422:
	case CAMWIRE_PIXEL_YUV444:
	default:
	    return(0);
    }
}

/*
  ----------------------------------------------------------------------
  Return the equivalent pixel depth with 8-bit components, given the
  camera pixel depth in bits.
*/
static int equivalent_8bit(const int cam_depth)
{
    switch (cam_depth)
    {
	case 16:
	case 48:
	    return(cam_depth/2);
	case 8:
	case 24:
	default:
	    return(cam_depth);
    }
}

/*
  ----------------------------------------------------------------------
  Create the SDL screen display surface.  Returns NULL on error.
*/
static SDL_Surface * create_display_surface(const int width,
					    const int height,
					    const int depth)
{
    Uint32 video_flags;
    SDL_Surface *screen;

    /* Set the video mode: */
    /* Note: X11 does not directly support video hardware acceleration,
       but may do so using OpenGL.
    video_flags = SDL_HWSURFACE | SDL_HWACCEL | SDL_DOUBLEBUF;
    */
    video_flags = SDL_SWSURFACE;
    screen = SDL_SetVideoMode(width, height, depth, video_flags);
    if (screen == NULL)
    {
	fprintf(stderr, "SDL_SetVideoMode() failed in "
		"create_display_surface():\n%s\n", SDL_GetError());
	return(NULL);
    }
    /*
    if ((screen->flags & SDL_HWSURFACE) == 0)
    {
	fprintf(stderr, "FYI: Could not use video hardware directly. SDL "
		"screen surface is in user memory.\n");
    }
    */

    /* Create a palette if necessary: */
    if (depth == 8)  set_linear_palette(screen);

    return(screen);
}

/*
  ----------------------------------------------------------------------
  Create an empty SDL frame surface.  Returns NULL on error.
*/
static SDL_Surface * create_frame_surface(const int width,
					  const int height,
					  const int depth)
{
    int pitch;
    Uint32 rmask, gmask, bmask, amask;
    SDL_Surface *frame;
    
    /* Set up the frame surface parameters: */
    pitch = width*depth/8;
    if (depth == 8)
    {
	rmask = 0;
	gmask = 0;
	bmask = 0;
	amask = 0;
    }
    else if (depth == 24)
    {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0xff000000;
	gmask = 0x00ff0000;
	bmask = 0x0000ff00;
	amask = 0;
#else
	rmask = 0x000000ff;
	gmask = 0x0000ff00;
	bmask = 0x00ff0000;
	amask = 0;
#endif
    }
    else
    {
	fprintf(stderr, "%s: %s(): Internal error: Display depth %d is not 8 or 24.\n",
		__FILE__, __func__, depth);
	return(NULL);
    }
    
    /* Create the frame surface: */
    frame = SDL_CreateRGBSurfaceFrom(NULL,
				       width, height, depth, pitch,
				       rmask, gmask, bmask, amask);
    if (frame == NULL)
    {
	fprintf(stderr, "SDL_CreateRGBSurfaceFrom() failed in %s():\n%s\n", __func__,
		SDL_GetError());
	return(NULL);
    }

    /* Create a palette if necessary: */
    if (depth == 8)  set_linear_palette(frame);

    return(frame);
}

/*
  ----------------------------------------------------------------------
  Fill in the 8-bit colour map:
*/
static void set_linear_palette(SDL_Surface *surface)
{
    SDL_Color map[256];
    int i;

    for (i = 0; i < 256; i++)  map[i].r = map[i].g = map[i].b = i;
    SDL_SetColors(surface, map, 0, 256);
}
