/***********************************************************************

    Copyright (c) Industrial Research Limited 2004

    This file is part of Camwire, a generic camera interface.

    Camwire is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2.1 of the
    License, or (at your option) any later version.

    Camwire is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Camwire; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA


    Title: Simple Xvideo display server
    Based on the libdc1394 example dc1394_multiview.c by Dan Dennedy

    Description:
    This module uses Xvideo to display images.  It only works well when
    only one process is displaying at a time.  When more than one
    process uses it simultaneously, the displayed images are mixed up.

***********************************************************************/

#include <netinet/in.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/Xvlib.h>
#include <X11/keysym.h>

#include "dc1394/conversions.h"  /* RGB2YUV */

#include "display-codes.h"
#include "display-xv.h"


/* ok the following constant should be by right included thru in
 * Xvlib.h, obtained from xvinfo */
/* 16 bits/pixel, 1 plane, YUV packed:*/
#ifndef XV_YUY2
#define XV_YUY2 0x32595559L
#endif
/* 12 bits/pixel, 3 planes, YUV planar:*/
#ifndef XV_YV12
#define XV_YV12 0x32315659L
#endif
/* 16 bits/pixel, 1 plane, YUV packed:*/
#ifndef XV_UYVY
#define XV_UYVY 0x59565955L
#endif
/* 12 bits/pixel, 3 planes, YUV planar:*/
#ifndef XV_I420
#define XV_I420 0x30323449L
#endif


/* The display info points to this struct: */
typedef struct display_info_data
{
    Display *display;
    Window window;
    long xformat;
    GC gc;
    XvPortID base_id;
    void *frame_buffer;
    size_t display_width, display_height;
    size_t device_width, device_height;
    Camwire_pixel device_coding;
    unsigned long maxval;
}
Display_info_data;


/* Prototypes */
static int set_frame_format(Display_info info,
			    const int cam_width,
			    const int cam_height,
			    const Camwire_pixel coding);
static void iyu12yuy2 (const unsigned char *src,
		       unsigned char *dest,
		       const size_t NumPixels);
static void rgb2yuy2 (const unsigned char *RGB,
		      unsigned char *YUV,
		      const size_t NumPixels);
static void rgb2yuy2_16(const uint16_t *src,
			uint8_t *dest,
			const size_t numpix,
			const unsigned long mv);
static void mono2yuy2(const unsigned char *src,
		      unsigned char *dest,
		      const size_t NumPixels);
static void mono2yuy2_16(const uint16_t *src,
			 uint8_t *dest,
			 const size_t numpix,
			 const unsigned long mv);
static void convert_frames(Display_info info, const void *capture_buffer);
static void display_frames(Display_info info);
static int QueryXv(Display_info info);


/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
Display_info display_xv_create(const int cam_width,
			       const int cam_height,
			       const Camwire_pixel coding,
			       const unsigned long maxval)
{
    Display_info info;
    XGCValues xgcv;
    unsigned long background = 0x010203L;

    info = (Display_info_data *) calloc(1, sizeof(Display_info_data));
    if (info == NULL)  return NULL;
    
    /* Force the frame buffer resize in set_frame_format(): */
    if (set_frame_format(info, cam_width, cam_height, coding) != DISPLAY_SUCCESS)
	return NULL;

    /* make the window */
    info->display = XOpenDisplay(getenv("DISPLAY"));
    if (info->display == NULL)
    {
	fprintf(stderr, "Could not open display \"%s\"\n",
	        getenv("DISPLAY"));
	display_xv_destroy(info);
	return NULL;
    }
    
    if (QueryXv(info) < 0)
    {
	display_xv_destroy(info);
	return NULL;
    }
    
    info->display_width = info->device_width;
    info->display_height = info->device_height;
    info->maxval = maxval;
    
    info->window =
	XCreateSimpleWindow(info->display,
			    DefaultRootWindow(info->display),
			    0, 0,
			    info->display_width,
			    info->display_height,
			    0,
			    WhitePixel(info->display,
				       DefaultScreen(info->display)),
			    background);
    
    XSelectInput(info->display, info->window,
		 StructureNotifyMask | KeyPressMask);
    XMapWindow(info->display, info->window);
    /* connection = ConnectionNumber(info->display); */
    (void) ConnectionNumber(info->display);
    
    info->gc = XCreateGC(info->display, info->window, 0, &xgcv);
    return info;
} /* display_xv_create() */

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
void display_xv_destroy(Display_info info)
{
    if ((void*) info->window !=  0)
    {
	XUnmapWindow(info->display, info->window);
    }
    info->window = 0;
    if (info->display != NULL) 	XFlush(info->display);
    free(info->frame_buffer );
    free(info);
} /* display_xv_destroy() */

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_xv_resize(Display_info info, const int cam_width,
		      const int cam_height)
{
    size_t newwidth, newheight;
    
    newwidth = cam_width*info->display_width/info->device_width;
    newheight = cam_height*info->display_height/info->device_height;
    if (set_frame_format(info, cam_width, cam_height, info->device_coding) !=
	DISPLAY_SUCCESS)
    {
	return DISPLAY_FAILURE;
    }
    info->display_width = newwidth;
    info->display_height = newheight;
    XResizeWindow(info->display, info->window,
		  info->display_width,info->display_height);
    return DISPLAY_SUCCESS;
} /* display_xv_resize() */

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_xv_coding(Display_info info, const Camwire_pixel coding)
{
    return set_frame_format(info, info->device_width, info->device_height, coding);
} /* display_xv_coding() */

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_xv_maxval(Display_info info, const unsigned long maxval)
{
    if (maxval > 0xffff)  return(DISPLAY_FAILURE);
    info->maxval = maxval;
    return(DISPLAY_SUCCESS);
} /* display_xv_maxval() */

/*
  ----------------------------------------------------------------------
  See display.h for documentation on this function.
*/
int display_xv_frame(Display_info info, const void *capture_buffer)
{
    XEvent xev;
    
    convert_frames(info, capture_buffer);
    display_frames(info);
    /*XFlush(info->display);*/
    
    while(XPending(info->display) > 0)
    {
	XNextEvent(info->display, &xev);
	switch (xev.type)
	{
	    case ConfigureNotify:
		info->display_width = xev.xconfigure.width;
		info->display_height = xev.xconfigure.height;
		display_frames(info);
		break;
	    case KeyPress:
		switch (XkbKeycodeToKeysym(info->display,
					   xev.xkey.keycode, 0, 0))
		{
		    case XK_q:
		    case XK_Q: 	/* Quit.*/
			/*display_xv_destroy();*/
			return DISPLAY_QUIT_REQ;
			break;
		    case XK_comma:
		    case XK_less:
			info->display_width = info->display_width/2;
			info->display_height =
			    info->display_height/2;
			XResizeWindow(info->display,
				      info->window,
				      info->display_width,
				      info->display_height);
			display_frames(info);
			break;
		    case XK_period:
		    case XK_greater:
			info->display_width = 2*info->display_width;
			info->display_height =
			    2*info->display_height;
			XResizeWindow(info->display,
				      info->window,
				      info->display_width,
				      info->display_height);
			display_frames(info);
			break;
		}
		break;
	} /* switch*/
    } /* XPending*/
    return DISPLAY_SUCCESS;
} /* display_xv_frame() */


/* helper functions */



static int set_frame_format(Display_info info,
			    const int cam_width,
			    const int cam_height,
			    const Camwire_pixel coding)
{
    switch(coding)
    {
	case CAMWIRE_PIXEL_YUV422:
	    info->xformat = XV_UYVY;
	    break;
	default:
	    info->xformat = XV_YUY2;
    }

    /* (Re-)allocate memory for display buffer (only if size has
       changed): */
    if ((size_t) cam_width*cam_height != info->device_width*info->device_height)
    {
	info->frame_buffer =
	    realloc(info->frame_buffer, (size_t) cam_width*cam_height*2);
    }
	
    info->device_width = cam_width;
    info->device_height = cam_height;
    info->device_coding = coding;
    return DISPLAY_SUCCESS;
} /* set_frame_format() */


/* image format conversion functions */

/* Convert 12 bits/pixel YUV411 into 16 bits/pixel YUY2: */
static void iyu12yuy2(const unsigned char *src,
		      unsigned char *dest,
		      const size_t numpix)
{
    unsigned char *d, *s_end, *s;
    int u, v;

    d = dest;
    s = (unsigned char *) src;
    s_end = (unsigned char *) src + numpix*3/2;
    while (s < s_end)
    {
	u = *s++;
	*d++ = *s++;  /* Y0 */
	*d++ = u;     /* U */
	*d++ = *s++;  /* Y1 */
	v  = *s++;
	*d++ = v;     /* V */
	*d++ = *s++;  /* Y2 */
	*d++ = u;     /* U */
	*d++ = *s++;  /* Y3 */
	*d++ = v;     /* V */
    }
} /* iyu12yuy2() */


/* Convert 24 bits/pixel RGB into 16 bits/pixel YUY2: */
static void rgb2yuy2(const unsigned char *src,
		     unsigned char *dest,
		     const size_t numpix)
{
    unsigned char *d, *s_end, *s;
    int y0, y1, u0, u1, v0, v1 ;
    int r, g, b;

    d = dest;
    s = (unsigned char *) src;
    s_end = (unsigned char *) src + 3*numpix;
    while (s < s_end)
    {
	r = *s++;
	g = *s++;
	b = *s++;
	RGB2YUV(r, g, b, y0, u0, v0);
	r = *s++;
	g = *s++;
	b = *s++;
	RGB2YUV(r, g, b, y1, u1, v1);
	*d++ = y0;
	*d++ = (u0 + u1)/2;
	*d++ = y1;
	*d++ = (v0 + v1)/2;
    }
} /* rgb2yuy2() */


/* Convert 48 bits/pixel RGB into 16 bits/pixel YUY2: */
static void rgb2yuy2_16(const uint16_t *src,
			uint8_t *dest,
			const size_t numpix,
			const unsigned long mv)
{
    unsigned long maxval;
    int shift;
    uint8_t *d;
    uint16_t *s, *s_end;
    int y0, y1, u0, u1, v0, v1 ;
    int r, g, b;

    /* Work out the bit shift corresponding to maxval: */
    maxval = mv;
    shift = 0;
    while (maxval != 0)
    {
	maxval >>= 1;
	++shift;
    }
    if (shift < 8 || shift > 16)  shift = 16;
    shift -= 8;

    /* Shift and copy: */
    d = dest;
    s = (uint16_t *) src;
    s_end = (uint16_t *) src + 3*numpix;
    while (s < s_end)
    {
	r = ntohs(*s++) >> shift;
	g = ntohs(*s++) >> shift;
	b = ntohs(*s++) >> shift;
	RGB2YUV(r, g, b, y0, u0, v0);
	r = ntohs(*s++) >> shift;
	g = ntohs(*s++) >> shift;
	b = ntohs(*s++) >> shift;
	RGB2YUV(r, g, b, y1, u1, v1);
	*d++ = y0;
	*d++ = (u0 + u1)/2;
	*d++ = y1;
	*d++ = (v0 + v1)/2;
    }
} /* rgb2yuy2_16() */


/* Convert 8 bits/pixel MONO into 16 bits/pixel YUY2: */
static void mono2yuy2(const unsigned char *src,
		      unsigned char *dest,
		      const size_t numpix)
{
    unsigned char *d, *s_end, *s;

    d = dest;
    s_end = (unsigned char *) src + numpix;
    for (s=(unsigned char *)src; s<s_end; ++s)
    {
	*d++ = *s;
	*d++ = 128;
    }
} /* mono2yuy2() */


/* Convert 16 bits/pixel MONO into 16 bits/pixel YUY2: */
static void mono2yuy2_16(const uint16_t *src,
			 uint8_t *dest,
			 const size_t numpix,
			 const unsigned long mv)
{
    unsigned long maxval;
    int shift;
    uint8_t *d;
    uint16_t *s, *s_end;

    /* Work out the bit shift corresponding to maxval: */
    maxval = mv;
    shift = 0;
    while (maxval != 0)
    {
	maxval >>= 1;
	++shift;
    }
    if (shift < 8 || shift > 16)  shift = 16;
    shift -= 8;

    /* Shift and copy: */
    d = dest;
    s_end = (uint16_t *) src + numpix;
    for (s=(uint16_t *)src; s<s_end; ++s)
    {
	*d++ = ntohs(*s) >> shift;
	/* *d++ = *s >> shift; */  /* For non-compliant little-endian pixel format. */
	*d++ = 128;
    }
} /* mono2yuy2() */


/* Convert all pixel types to XV's YUY2 format (equivalent to YUV422): */
static void convert_frames(Display_info info, const void *capture_buffer)
{
    size_t numpix;

    numpix = info->device_width*info->device_height;
    switch (info->device_coding)
    {
	case CAMWIRE_PIXEL_MONO8:  /* 8 bits per pixel.*/
	case CAMWIRE_PIXEL_RAW8:
	    mono2yuy2(capture_buffer, info->frame_buffer, numpix);
	    break;

	case CAMWIRE_PIXEL_YUV411:  /* 12 bits per pixel.*/
	    iyu12yuy2(capture_buffer, info->frame_buffer, numpix);
	    break;

	case CAMWIRE_PIXEL_YUV422:  /* 16 bits per pixel.*/
	    memcpy(info->frame_buffer, capture_buffer, numpix*2);
	    break;

	case CAMWIRE_PIXEL_MONO16:  /* 16 bits per pixel.*/
	case CAMWIRE_PIXEL_MONO16S:
	case CAMWIRE_PIXEL_RAW16:
	    mono2yuy2_16(capture_buffer, info->frame_buffer, numpix, info->maxval);
	    break;

	case CAMWIRE_PIXEL_YUV444:  /* 24 bits per pixel.*/
	case CAMWIRE_PIXEL_RGB8:
	    rgb2yuy2(capture_buffer, info->frame_buffer, numpix);
	    break;
	    
	case CAMWIRE_PIXEL_RGB16:  /* 48 bits per pixel.*/
	case CAMWIRE_PIXEL_RGB16S:
	    rgb2yuy2_16(capture_buffer, info->frame_buffer, numpix, info->maxval);
	    break;
	    
	case CAMWIRE_PIXEL_INVALID:
	default:
	    /* Do nothing. */
	    break;
    }
} /* convert_frames() */


static void display_frames(Display_info info)
{
    XvImage *xv_image;

    xv_image = XvCreateImage(info->display,
			     info->base_id,
			     (int) info->xformat,
			     (char *) info->frame_buffer,
			     (int) info->device_width,
			     (int) info->device_height);
    XvPutImage(info->display,
	       info->base_id,
	       info->window,
	       info->gc,
	       xv_image,
	       0, 0,
	       info->device_width, info->device_height,
	       0, 0,
	       info->display_width, info->display_height);
} /* display_frames() */


static int QueryXv(Display_info info)
{
    unsigned int i, num_adaptors;
    int adaptor;
    XvAdaptorInfo *xv_info;
    int j, num_xformats;
    XvImageFormatValues *xformats = 0;
    /*char xv_name[5];*/
    
    XvQueryAdaptors(info->display,
		    DefaultRootWindow(info->display),
		    &num_adaptors,
		    &xv_info);

    adaptor = -1;
    for (i = 0; i < num_adaptors; i++) {
	xformats = XvListImageFormats(info->display,
				      xv_info[i].base_id,
				      &num_xformats);
	for (j = 0; j < num_xformats; j++) {
	    if (xformats[j].id == info->xformat) {
		/*
		xv_name[4] = 0;
		memcpy(xv_name, &xformats[j].id, 4);
		fprintf(stderr, "using Xv format 0x%x %s %s\n",
		        xformats[j].id,
			xv_name,
			(xformats[j].format == XvPacked) ? "packed" :
		                                           "planar");
		*/
		if (adaptor < 0)  adaptor = i;
		break;
	    }
	}
	XFree(xformats);
	if (adaptor >= 0)  break;
    }
    if (adaptor >= 0)  info->base_id = xv_info[adaptor].base_id;

    XvFreeAdaptorInfo(xv_info);

    if (adaptor < 0) {
	fprintf(stderr, "Could not find a suitable Xv adaptor.\n");
    }
    return adaptor;
} /* QueryXv() */
