/***********************************************************************

    Copyright (c) Industrial Research Limited 2006-2012

    This file is part of Camwire, a generic camera interface.

    Camwire is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2.1 of the
    License, or (at your option) any later version.

    Camwire is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Camwire; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA

    Title: Measure config

    Description:

    Attempts to use the computer's serial port to measure the timestamp
    parameters of IEEE 1394 cameras with external triggers.  If the
    camera does not appear to respond to the serial port, it falls back
    to using internal triggers which are not as accurate but gives
    approximately correct results.

    Measures the transmit_overlap, bus_speed, line_transfer_time and
    transmit_setup_time parameters of the camera timestamp calculation
    using a robust algorithm with minimal number of timestamp
    measurements.  Because it uses Camwire, it needs an existing .conf
    file for the camera (see camwire/camwire.h for details).  All values
    in the .conf file that are consistent with measurement are retained,
    and new measured values are substituted where appropriate.  The
    measured configuration is printed to stdout in Camwire .conf file
    format.

    Note that the measured transmit_setup_time excludes any
    trig_setup_time found in the Camwire .conf file.  If you don't know
    the trig_setup_time (the delay between the trigger and the start of
    integration) then it is probably safe to set a trig_setup_time of 0
    in the .conf file, in which case any actual trigger delay will be
    absorbed into the transmit_setup_time measurement.

    Overlapped transmission means that transmission to the host happens
    simultaneously with sensor read-out.  Non-overlapped transmission
    means that transmission to the host does not start until the whole
    sensor has been read out to an internal frame buffer.

    Let p be the number of packets per frame and h be the image height.
    Both have camera-imposed limits p_min, p_max, h_min and h_max.

    Total non-overlapped transmission time:

      T_total(p,h) = transmit_setup_time + bus_period*p + line_transfer_time*h

    Total overlapped transmission time:

      T_total(p,h) =
	transmit_setup_time + max(bus_period*p, line_transfer_time*h)

    where bus_period = 0.05/bus_speed (bus_speed in 100, 200, 400...).

    The following table summarizes the rough shape of the total transmission
    time function T_total(p,h):

      point            T_total
      (p_min,h_min)    minimum
      (p_min,h_max)    dominated by sensor line transfer time
      (p_max,h_min)    dominated by image bus transmission time
      (p_max,h_max)    maximum
      (in between)     if overlapped, bi-planar, else planar

    p is set in practice by adjusting the framerate in Format 7.

    If the measurements are done with the camera NOT set to Format 7,
    the line_transfer_time will come out to 0, which is fine as long as
    the camera is always used in the same Format.  If you intend
    switching the camera between different Formats, then the
    configuration is best measured in Format 7.  If the camera does not
    support Format 7 and you need accurate timestamps, then you will
    need to create and use a separate configuration file for each
    Format.

    Johann Schoonees
    $Id: measureconf_1394.c,v 20.5 2012/01/10 22:20:53 johanns Exp $

***********************************************************************/

/*
  ----------------------------------------------------------------------
  Includes:
*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "dc1394/dc1394.h"

#include "camwire/camwire_handle.h"
#include "camwire/camwirebus.h"
#include "camwire/camwire.h"

/*
  ----------------------------------------------------------------------
  Defines:
*/
/* Max number of chars in serial device name: */
#define MAX_DEV_NAME 100
/* Max number of chars in error message string: */
#define MAX_ERR_MSG 200
#define ERROR_IF_CAMWIRE_FAIL(r) \
    if ((r)!=CAMWIRE_SUCCESS) \
      {fprintf(stderr,"In %s, line %d: ", __FILE__, __LINE__); \
       errorexit(c_handle, __func__, "Camwire function call failed.");}
#define ERROR_IF_DC1394_FAIL(r)	\
    if ((r)!=DC1394_SUCCESS) \
      {fprintf(stderr,"In %s, line %d: ", __FILE__, __LINE__); \
       errorexit(c_handle, __func__, "dc1394 function call failed.");}

/*
  ----------------------------------------------------------------------
  Global types:
*/
/* Structure for holding accumulated linear regression data: */
typedef struct linreg
{
    double sum_x, sum_y, sum_x2, sum_xy;
    int N;
} Linreg;

/*
  ----------------------------------------------------------------------
  Prototypes:
*/
/* Housekeeping: */
static int process_command_line(int argc, char *argv[], char *serial_dev,
				double *meas_duration);
static void print_usage(const char *argv0);
static void print_description();

/* Principal function: */
static void find_time_params(const Camwire_handle c_handle,
			     const int ext_trig_OK,
			     const int serial_dev);

/* Camera functions: */
static Camwire_handle initialize_camera();
static int setup_trigger(const Camwire_handle c_handle, const int serial_dev);
static void get_cam_limits(const Camwire_handle c_handle,
			   int *width_unit, int *height_unit,
			   int *width_min, int *width_max,
			   int *height_min, int *height_max,
			   int *packets_min, int *packets_max);
static double T_total(const Camwire_handle c_handle,
		      const int p,
		      const int h,
		      const int w,
		      const int M,
		      const int ext_trig_OK,
		      const int serial_dev);
static double set_random_shutter(const Camwire_handle c_handle);
static double get_image_timestamp(const Camwire_handle c_handle);
static double assumed_transmit_time(const Camwire_handle c_handle,
				    const int num_packets, const int height);
static void errorexit(const Camwire_handle c_handle, const char *funcname, const char *msg);
static void shutdown(const Camwire_handle c_handle);
static void complete_acquisition(const Camwire_handle c_handle);
static double numpackets2framerate(const int num_packets);
static int framerate2numpackets(const double frame_rate);

/* Serial port: */
static double trigger(const Camwire_handle c_handle, const int serial_dev);
static double soft_trigger(const Camwire_handle c_handle);
static void setDTR(const int serial_dev, const int state);

/* Helper functions: */
static inline double bus_period(const int speed);
static inline int ceil_ratio(const int num, const int denom);
static inline double max(const double a, const double b);
static inline double min(const double a, const double b);
static inline int xor(const int a, const int b);
static void acc(const double h, const double T, Linreg *rg);
static inline double uniform();
static int choose_M(const int p);
inline static dc1394video_mode_t convert_format_mode2dc1394video_mode(const int format,
								      const int mode);
/* inline static int fixed_image_size(const dc1394video_mode_t video_mode); */
inline static int variable_image_size(const dc1394video_mode_t video_mode);


/*
  ----------------------------------------------------------------------
  Global variables:
*/
Camwire_conf config;  /* Our local copy of the camera configuration.*/
double measurement_duration;  /* The maximum time allowed for measuring
				 a transmission delay.  Set in main()
				 and used in choose_M().*/
/* FIXME: the single_shot_capable flag is already present in
   Camwire_handle->User_handle->extras: */
int single_shot_capable;  /* Flag which is true if camera has single-shot mode.*/

/*
  ----------------------------------------------------------------------
  Main:
 */
int main(int argc, char *argv[])
{
    int run_tests;
    char serial_dev_name[MAX_DEV_NAME+1];
    unsigned int seed;
    int serial_dev;
    char err_msg[MAX_ERR_MSG+1];
    Camwire_handle c_handle;
    int external_trigger_OK;
    
    run_tests = process_command_line(argc, argv, serial_dev_name,
				     &measurement_duration);
    if (!run_tests)  return 0;

    seed = time(0) - 1111111111L;
    srand(seed);  /* For rand() in uniform(). */

    /* Initialize the serial port: */
    serial_dev = open(serial_dev_name, O_RDWR);
    if (serial_dev < 0)
    {
	snprintf(err_msg, MAX_ERR_MSG, "Could not open the serial device %s",
		 serial_dev_name);
	errorexit(0, __func__, err_msg);
    }
    setDTR(serial_dev, 1);  /* Set high to start */

    /* Initialize the camera and its configuration parameters: */
    c_handle = initialize_camera();  /* Also initializes global variable config.*/
    external_trigger_OK = setup_trigger(c_handle, serial_dev);
    if (!external_trigger_OK)
	fprintf(stderr,
		"Using internal trigger: results will be approximate.\n");

    /* Measure and print the timestamp parameters: */
    find_time_params(c_handle, external_trigger_OK, serial_dev);
    camwire_write_config_to_file(stdout, &config);  /* Same format as conf file. */
    fflush(stdout);

    /* Wrap up: */
    shutdown(c_handle);
    return 0;
} /* main() */

/*
  ----------------------------------------------------------------------
  Initialize variables from the command line options.
 */
static int process_command_line(int argc, char *argv[], char *serial_dev,
				double *meas_duration)
{
    int run_tests;
    int c;

    /* Defaults: */
    run_tests = 1;
    strncpy(serial_dev, "/dev/ttyS0", MAX_DEV_NAME);
    *meas_duration = 10.0;
    
    while(1) {
	c = getopt(argc, argv, "hds:m:");
	if (c == -1)
	    break;

	switch (c) {
	    case 'h':
		print_usage(argv[0]);
		run_tests = 0;
		break;
		
	    case 'd':
		print_description();
		run_tests = 0;
		break;
		
	    case 's':
		strncpy(serial_dev, optarg, MAX_DEV_NAME);
		break;
		
	    case 'm':
		*meas_duration = atof(optarg);
		if (*meas_duration < 0.1)
		    *meas_duration = 0.1;  /* Guard.*/
		break;
		
	    case '?':
		print_usage(argv[0]);
		run_tests = 0;
		break;

	    default:
		fprintf(stderr, "Internal error: getopt() returned 0%o.\n", c);
		run_tests = 0;
	}
	if (!run_tests)  break;
    }

    if (run_tests && optind < argc) {
	fprintf(stderr, "%s: invalid argument --", argv[0]);
	while (optind < argc)
	    fprintf(stderr, " %s", argv[optind++]);
	fprintf(stderr, "\n");
	run_tests = 0;
    }

    return run_tests;
} /* process_command_line() */

/*
  ----------------------------------------------------------------------
  Prints the command-line invocation.
 */
static void print_usage(const char *argv0)
{
    fprintf(stderr,
	"\n"
	"Usage: %s [-h] [-d] [-s SERIAL_DEVICE_NAME] [-m MEASUREMENT_DURATION]\n",
	argv0);
    fprintf(stderr,
	    "\n"
	    "\t-h : prints this help.\n"
	    "\t-d : prints a more detailed description.\n"
	    "\t-s : serial port device name (default \"/dev/ttyS0\").\n"
	    "\t-m : measurement duration in seconds (default 10.0).\n"
	    );
} /* print_usage() */

/*
  ----------------------------------------------------------------------
  Prints a program description.
 */
static void print_description()
{
    fprintf(stderr,
"\n"
"    Description:\n"
"\n"
"    Measures the parameters needed for accurate timestamp calculation\n"
"    for IEEE 1394 cameras.  This can be done to an accuracy of a few\n"
"    microceconds if the camera has an external trigger input connected\n"
"    to the DTR line of the computer's serial port (SERIAL_DEVICE_NAME).\n"
"    If the camera does not have an external trigger or you want to make\n"
"    do without constructing the necessary circuitry for connecting to\n"
"    the serial port (see below), then it falls back to using internal\n"
"    triggers which are not as accurate but gives approximately correct\n"
"    results.\n"
"\n"
"    The program does seven measurements (each of about\n"
"    MEASUREMENT_DURATION seconds) to calculate the transmit_overlap,\n"
"    bus_speed, line_transfer_time and transmit_setup_time parameters\n"
"    and print them to stdout in Camwire configuration (.conf) file\n"
"    format.  Because it uses Camwire, it needs an existing .conf file\n"
"    for the camera (see camwire/camwire.h for details).  All values in\n"
"    the .conf file that are consistent with measurement are retained,\n"
"    and new measured values are substituted where appropriate.\n"
"\n"
"    If the measurements are done with the camera NOT set to Format 7,\n"
"    the line_transfer_time will come out to 0, which is fine as long as\n"
"    the camera is always used in the same Format.  If you intend\n"
"    switching the camera between different Formats, then the\n"
"    configuration is best measured in Format 7.  If the camera does not\n"
"    support Format 7 and you need accurate timestamps, then you will\n"
"    need to create and use a separate configuration file for each\n"
"    Format.\n"
"\n"
"    Note that the measured transmit_setup_time excludes any\n"
"    trig_setup_time found in the Camwire .conf file.  If you don't know\n"
"    the trig_setup_time (the delay between the trigger and the start of\n"
"    integration) then it is probably safe to set a trig_setup_time of 0\n"
"    in the .conf file, in which case any actual trigger delay will be\n"
"    absorbed into the transmit_setup_time measurement.\n"
"\n"
"    I use the following circuit to provide a trigger pulse from the\n"
"    serial port to my camera's external trigger input.  It provides\n"
"    TTL-level (0 to +5V) signals from a serial port with RS-232-level\n"
"    outputs (-12V to +12V) which may be the most common situation.  IT\n"
"    IS YOUR OWN RESPONSIBILITY TO ENSURE THAT YOUR CAMERA WON'T BE\n"
"    DAMAGED BY ANYTHING YOU CONNECT TO IT.  Since I cannot know what\n"
"    serial port and what camera you are using, the information provided\n"
"    here is without any implied fitness for your purpose.  It may be a\n"
"    good idea to use a voltmeter or oscilloscope to check that the\n"
"    voltage swing and current is within the specified range for your\n"
"    camera's trigger input.\n"
"\n"
"    serial pin 4 DTR --- R ---------- to camera external trigger input\n"
"                              |  |-\n"
"      	                       R  D\n"
"                              |  |+\n"
"    serial pin 5 GND -------------\n"
"\n"
"    where each R is a 470 ohm resistor and D is a signal diode with its\n"
"    anode (+) connected to GND and its cathode (-) connected to the\n"
"    camera input.  The serial port pin numbers refer to a standard 9-pin\n"
"    serial port connector.  Other types of connectors would have\n"
"    different pin-outs.\n"
	    );
} /* print_description() */

/*
  ----------------------------------------------------------------------
  Do the time parameter estimation: transmit_overlap, bus_speed,
  line_transfer_time and transmit_setup_time.  These members of
  Camwire_conf contain initial values assigned by Camwire, and contain
  measured values on return.
 */
static void find_time_params(const Camwire_handle c_handle,
			     const int ext_trig_OK,
			     const int serial_dev)
{
    int w_unit, h_unit;
    int w_max, h_max, w_min, h_min, p_min, p_max;
    int M, nearest_valid;
    int few, i, h;
    double T0, T1, T2, T3;
    double bus_freq, logspeed;
    double T3_predict_non_overlap, T3_predict_overlap;
    double h_inc, T, denom;
    Linreg rg;
    int b_dummy;

    /* Discover good camera and image settings for the tests: */
    get_cam_limits(c_handle, &w_unit, &h_unit, &w_min, &w_max, &h_min, &h_max,
		   &p_min, &p_max);
    
    /* Transmission time at point (p_min,h_min).  This is presumably at the
       T_total function minimum: */
    M = choose_M(p_min);
    T0 = T_total(c_handle, p_min, h_min, w_min, M, ext_trig_OK, serial_dev);

    /* Transmission time at point (p_max,h_min).  This is presumably dominated
       by image bus transmission time: */
    M = choose_M(p_max);
    T1 = T_total(c_handle, p_max, h_min, w_max, M, ext_trig_OK, serial_dev);
    if (T1 <= T0)  errorexit(c_handle, __func__, "Could not determine the bus speed.");

    /* Quantize bus_speed to nearest valid value: */
    bus_freq = (p_max - p_min)/(T1 - T0);
    logspeed = log(1e-3*bus_freq)/log(2);  /* 1=100Mb/s, 2=200Mb/s,
							 3=400Mb/s etc.*/
    nearest_valid = (int)(logspeed + 0.5);  /* Round.*/
    if (fabs(logspeed - nearest_valid) > 0.2)
	errorexit(c_handle, __func__, "Could not determine bus speed.");
    b_dummy = (int)(50*pow(2, nearest_valid) + 0.5);
    if (b_dummy != config.bus_speed)
	errorexit(c_handle, __func__, "Measured bus speed different to config.");

    /* Transmission time at point (p_min,h_max).  This is presumably dominated
       by sensor line transfer time: */
    M = choose_M(p_min);
    T2 = T_total(c_handle, p_min, h_max, w_min, M, ext_trig_OK, serial_dev);

    /* Transmission time at point (p_max,h_max).  This is presumably at the
       T_total function maximum: */
    M = choose_M(p_max);
    T3 = T_total(c_handle, p_max, h_max, w_max, M, ext_trig_OK, serial_dev);

    /* Decide whether transmission is overlapped or not: */
    T3_predict_non_overlap = T1 + T2 - T0;
    T3_predict_overlap = max(T1, T2);
    config.transmit_overlap =
	(int)(T3 < (T3_predict_overlap + T3_predict_non_overlap)/2);

    /* Estimate line_transfer_time and transmit_setup_time: */

    /* Set up linear regression, including measurements already done: */
    rg.sum_x = rg.sum_y = rg.sum_x2 = rg.sum_xy = 0.0;
    rg.N = 0;
    /* Don't include T0 if overlapped: */
    if (!config.transmit_overlap)  acc(h_min, T0, &rg);
    acc(h_max, T2, &rg);

    /* Measure a few more times along p_min: */
    few = 4;
    h_inc = (double)(h_max - h_min)/few;
    M = choose_M(p_min);

    for (i=1; i<few; ++i)
    {
	h = (int)(h_min + i*h_inc + 0.5);  /* Round.*/
	h = ceil_ratio(h, h_unit)*h_unit;  /* Multiple of unit.*/
	T = T_total(c_handle, p_min, h, w_min, M, ext_trig_OK, serial_dev);
	acc(h, T, &rg);
    }

    /* Final values of line_transfer_time and transmit_setup_time: */
    denom = rg.N*rg.sum_x2 - rg.sum_x*rg.sum_x;
    if (fabs(denom) > 1e-08)
	config.line_transfer_time = (rg.N*rg.sum_xy - rg.sum_y*rg.sum_x)/denom;
    else
	config.line_transfer_time = 0;
    config.transmit_setup_time = (rg.sum_y - config.line_transfer_time*rg.sum_x)/rg.N;
    if (!config.transmit_overlap)
	config.transmit_setup_time -= bus_period(config.bus_speed)*p_min;

} /* find_time_params() */

/*
  ----------------------------------------------------------------------
  Set up the camera under test.  There should be only one camera
  connected to the bus because we want to use all the available
  bandwidth:
 */
static Camwire_handle initialize_camera()
{
    int num_cameras, num_buffers;
    Camwire_handle *handle_array;
    Camwire_handle c_handle;
    int result;
    struct timespec snooze;
    dc1394video_mode_t video_mode;
    uint32_t depth, c;
    dc1394color_coding_t color_id;
    dc1394color_codings_t coding_list;
    
    /* Initialize a camera: */
    handle_array = camwire_bus_create(&num_cameras);
    if (num_cameras <= 0 || !handle_array)
	errorexit(0, __func__, "Could not initialize the camera bus.");
    if (num_cameras > 1)
    {
	fprintf(stderr, "Found %d cameras.  ", num_cameras);
	errorexit(0, __func__, "This test works with a single camera only.");
    }
    c_handle = handle_array[0];
    result = camwire_create(c_handle);
    if (result != CAMWIRE_SUCCESS)
	errorexit(0, __func__, "Could not initialize the camera.");

    /* Make sure it's stopped: */
    result = camwire_set_run_stop(c_handle, 0);
    if (result != CAMWIRE_SUCCESS)
	errorexit(c_handle, __func__, "Could not stop the camera.");
    snooze.tv_sec = 0;
    snooze.tv_nsec = 500000000;  /* 0.5 s.*/
    nanosleep(&snooze, 0);  /* complete_acquisition() needs config.*/
    camwire_get_num_framebuffers(c_handle, &num_buffers);
    camwire_flush_framebuffers(c_handle, num_buffers, 0, 0);

    /* Initialize our copy of config: */
    if (camwire_get_config(c_handle, &config) != CAMWIRE_SUCCESS)
	errorexit(c_handle, __func__, "Could not obtain a camera configuration.");
    config.transmit_overlap = 0;  /* Override camera to ensure
				     complete_acquisition() in
				     setup_trigger() works reliably. */
    if (config.max_packets == 0)  config.max_packets = 4095;  /* Safe default.*/

    video_mode = convert_format_mode2dc1394video_mode(config.format, config.mode);
    
    /* Make sure its pixel coding is 8-bit: */
    ERROR_IF_DC1394_FAIL(
	dc1394_get_color_coding_from_video_mode(camwire_handle_get_camera(c_handle),
						video_mode,
						&color_id));
    ERROR_IF_DC1394_FAIL(
	dc1394_get_color_coding_bit_size(color_id, &depth));
    if (depth != 8)
    {
	/* Pixel coding is not 8 bit, but in Format 7 we may change color_id:*/
	if (variable_image_size(video_mode))  /* Format 7.*/
	{
	    ERROR_IF_DC1394_FAIL(
		dc1394_format7_get_color_codings(
		    camwire_handle_get_camera(c_handle),
		    video_mode,
		    &coding_list));
	    for (c=0; c<coding_list.num; ++c)
	    {
		color_id = coding_list.codings[c];
		ERROR_IF_DC1394_FAIL(
		    dc1394_get_color_coding_bit_size(color_id, &depth));
		if (depth == 8)
		{
		    ERROR_IF_DC1394_FAIL(
			dc1394_format7_set_color_coding(
			    camwire_handle_get_camera(c_handle),
			    video_mode,
			    color_id));
		    break;
		}
	    }
	}
    }
    if (depth != 8)
    {
	errorexit(c_handle, __func__, "This test works only with cameras that "
		  "have an 8-bit pixel coding available.");
    }
    
    return c_handle;
} /* initialize_camera() */

/*
  ----------------------------------------------------------------------
  Set up triggering.  If the camera does not respond to external
  triggers sent out the serial port, then we prepare to use single-shot
  frames instead.
 */
static int setup_trigger(const Camwire_handle c_handle, const int serial_dev)
{
    int result, ext_trig;
    /* double trig_time; */
    void *buffer;

    single_shot_capable = 1;  /* True unless we find out otherwise.*/
    
    /* Find out if we can use an external trigger: */
    result = camwire_set_trigger_source(c_handle, 1);
    ext_trig = (result == CAMWIRE_SUCCESS);
    if (ext_trig)
    {
	camwire_set_trigger_polarity(c_handle, 1);
	camwire_set_run_stop(c_handle, 1);  /* Camera waits for a trigger.*/
	/* trig_time = */
	trigger(c_handle, serial_dev);
	complete_acquisition(c_handle);  /* Make sure it's arrived.*/
	result = camwire_point_next_frame_poll(c_handle, &buffer, 0);
	if (result != CAMWIRE_SUCCESS)
	    errorexit(0, __func__, "Could not capture a test frame.");
	if (buffer)
	{
	    fprintf(stderr, "Camera responds to trigger from serial port.\n");
	    camwire_unpoint_frame(c_handle);  /* All's well.*/
	}
	else
	{
	    fprintf(stderr,
		    "Camera did not respond to trigger from serial port.\n");
	    ext_trig = 0;
	}
    }
    else
    {
	fprintf(stderr, "Camera may not have an external trigger.\n");
    }
    if (!ext_trig)
    {
	/* Prepare to trigger camera by software: */
	camwire_set_run_stop(c_handle, 0);
	camwire_set_trigger_source(c_handle, 0);
	result = camwire_set_single_shot(c_handle, 1);
	single_shot_capable = (result == CAMWIRE_SUCCESS);
/* 	fprintf(stderr, "***single_shot_capable = %d\n", single_shot_capable);  // Debugging. */
    }
    return ext_trig;
} /* setup_trigger() */

/*
  ----------------------------------------------------------------------
  Get the minimum and maximum width, height and number of packets for
  the camera.

  The algorithm below is based on the equation the camera has to
  calculate internally:

      p = ceil(w h/s) = ceil(w_unit W h_unit H/(s_unit S))

  which implies

      s = ceil(w h/p)

  where w = w_unit W is the width,
        h = h_unit H is the height,
	s = s_unit S is the packet size in bytes,
	p is the number of packets per frame,
  and all variables are positive integers.

  We need to be able to keep p constant as H is varied arbitrarily.
  For each fixed p (we are interested in 2 of them, p_min and p_max), we
  can choose one fixed W:

      Choose W so that w_unit W h_unit/s_unit (the camera equation with
      H/S removed) is integer.  p is then chosen as the positive integer
      that makes

      W = s_unit/(w_unit h_unit) p

      an integer. p_min and W_min are found as the smallest such
      integers, subject to measured camera W_min.  p_max and W_max are
      the largest such integers, subject to measured camera p_max and
      W_max.

  For each H, we can choose a corresponding S:

      Choosing S = H is the only practical choice that always makes H/S
      integer if H can be any positive integer.  (Having S = w_unit H or
      S = h_unit H also works but will likely exceed the camera's
      S_max.)

  Reduce H_max if it exceeds S_max.  S_max may be a function of the
  image height in cameras that have a slow sensor read-out rate.

  Image size is adjusted directly through dc1394 in this function.
  Attempts to acquire images in this function will therefore probably
  result in segmentation faults because the camera DMA buffers will be
  the wrong size.
*/
static void get_cam_limits(const Camwire_handle c_handle,
			   int *w_unit, int *h_unit,
			   int *w_min, int *w_max,
			   int *h_min, int *h_max,
			   int *p_min, int *p_max)
{
    dc1394video_mode_t video_mode;
    int sq;
    int W, S;  /* Width and packetsize normalized to units.*/
    int s_max, s_unit, n_unit;
    double fps;
    int W_max, H_max;
    int s, p;  /* Unnormalized temporary variables.*/
    int su, ss, pp;  /* Temporaries.*/

    /* Ensure that shutter speed does not get in the way: */
    camwire_set_shutter(c_handle, 0.0);

    video_mode = convert_format_mode2dc1394video_mode(config.format, config.mode);
    
    /* Find 1st estimates for min and max image width and height: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_set_image_position(
		camwire_handle_get_camera(c_handle),
		video_mode,
		0, 0));
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_get_max_image_size(
		camwire_handle_get_camera(c_handle),
		video_mode,
		(uint32_t *)w_max, (uint32_t *)h_max));
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_get_unit_size(
		camwire_handle_get_camera(c_handle),
		video_mode,
		(uint32_t *)w_unit, (uint32_t *)h_unit));

	/* Take into account config.min_pixels: */
	if (config.min_pixels > 0)
	{
	    *w_min = ceil_ratio(config.min_pixels, (*w_unit)*(*h_max))*(*w_unit);
	    *h_min = ceil_ratio(config.min_pixels, (*h_unit)*(*w_max))*(*h_unit);
	    sq = sqrt(config.min_pixels);
	    if (*w_min < sq)  *w_min = ceil_ratio(sq, *w_unit)*(*w_unit);
	    if (*h_min < sq)  *h_min = ceil_ratio(sq, *h_unit)*(*h_unit);
	}
	else
	{
	    *w_min = *w_unit;
	    *h_min = *h_unit;
	}
    }
    else /* Other formats.*/
    {
	ERROR_IF_CAMWIRE_FAIL(
	    camwire_get_frame_size(c_handle, w_max, h_max));
	*w_min = *w_max;
	*h_min = *h_max;
	*w_unit = 1;
	*h_unit = 1;
    }
/*     fprintf(stderr, "****w_min = %d, *w_max = %d, *h_min = %d, *h_max = %d, *w_unit = %d, *h_unit = %d\n", *w_min, *w_max, *h_min, *h_max, *w_unit, *h_unit);  // Debugging. */

    /* Find packet unit size s_unit and max packet size s_max: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_set_image_size(
		camwire_handle_get_camera(c_handle),
		video_mode,
		*w_max, *h_max));
	/* PACKET_PARA_INQ is now valid. */
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_get_packet_parameters(camwire_handle_get_camera(c_handle),
						 video_mode,
						 (uint32_t *)&s_unit,
						 (uint32_t *)&s_max));
	*p_min = 0;  /* Avoid compiler warning about uninitialized variables.*/
    }
    else  /* Other formats.*/
    {
	s_unit = 4;
	ERROR_IF_CAMWIRE_FAIL(
	    camwire_set_framerate(c_handle, sqrt(DBL_MAX)));
	ERROR_IF_CAMWIRE_FAIL(
	    camwire_get_framerate(c_handle, &fps));
	*p_min = (int)(7200.0/fps + 0.5);  /* Find p_min at the same time.*/
	s_max = ceil_ratio((*w_max)*(*h_max), s_unit*(*p_min))*s_unit;
    }
/*     fprintf(stderr, "***s_unit = %d, s_max = %d\n", s_unit, s_max);  // Debugging. */

    /* Useful in the rest of the function: */
    W_max = (*w_max)/(*w_unit);
    H_max = (*h_max)/(*h_unit);
    n_unit = (*w_unit)*(*h_unit);
    
    /* Reduce max height h_max if it will be limited by s_max: */
    S = s_max/s_unit;
    if (H_max > S)
    {
	if (variable_image_size(video_mode))  /* Format 7.*/
	{
	    H_max = S;
	    *h_max = (*h_unit)*H_max;
	}
    }
/*     fprintf(stderr, "****h_max = %d\n", *h_max);  // Debugging. */

    /* Find 1st estimate for max number of packets p_max: */
    *p_max = config.max_packets;
    /* Limit imposed by popular Texas Instruments link layer controller chip:*/
    if (*p_max == 0 || *p_max > 4095)  *p_max = 4095;
/*     fprintf(stderr, "****p_max = %d\n", *p_max);  // Debugging. */

    /* Find the max packets p_max and associated w_max to use with it: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	pp = n_unit*W_max/s_unit;  /* Start from measured w_max.*/
	if (pp > *p_max)  pp = *p_max;
	W = s_unit*pp/n_unit;
	p = n_unit*W/s_unit;  /* Candidate p_max.*/
	while (p != pp)  /* Find largest integer solution. */
	{
	    pp = p;
	    W = s_unit*pp/n_unit;
	    p = n_unit*W/s_unit;
	}
	W_max = W;
	*w_max = (*w_unit)*W_max;
	*p_max = p;
    }
    else
    {
	ERROR_IF_CAMWIRE_FAIL(
	    camwire_set_framerate(c_handle, 0.0));
	ERROR_IF_CAMWIRE_FAIL(
	    camwire_get_framerate(c_handle, &fps));
	p = (int)(7200.0/fps + 0.5);
	if (p > *p_max)
	    errorexit(c_handle, __func__, "Could not set a usable framerate.");
	*p_max = p;
    }
/*     fprintf(stderr, "****p_max = %d, *w_max = %d\n", *p_max, *w_max);  // Debugging. */

    /* Find the min packets p_min and associated w_min to use with it: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	W = (*w_min)/(*w_unit);  /* Start from measured w_min.*/
	pp = ceil_ratio(n_unit*W, s_unit);
	if (pp < 2)  pp = 2;  /* Some cameras can't go too fast.*/
	W = ceil_ratio(s_unit*pp, n_unit);
	p = ceil_ratio(n_unit*W, s_unit);  /* Candidate p_min.*/
	while (p != pp)  /* Find smallest integer solution. */
	{
	    pp = p;
	    W = ceil_ratio(s_unit*pp, n_unit);
	    p = ceil_ratio(n_unit*W, s_unit);
	}
	*w_min = (*w_unit)*W;
	*p_min = p;
    }
    /* else */
    /* p_min was found in s_unit-s_max block above.*/
/*     fprintf(stderr, "****p_min = %d, *w_min = %d\n", *p_min, *w_min);  // Debugging. */

    /* Some cameras limit p_min (by limiting s_max) for large height,
       because of slow sensor read-out rates: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_set_image_size(
		camwire_handle_get_camera(c_handle),
		video_mode,
		*w_min, *h_max));
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_get_packet_parameters(camwire_handle_get_camera(c_handle),
						 video_mode,
						 (uint32_t *)&su, (uint32_t *)&ss));
	if (su != s_unit)
	    errorexit(c_handle, __func__, "Unit packet size has changed.");
	if (ss > s_max)
	    errorexit(c_handle, __func__, "Internal error: Bad s_max estimate.");
	
	s = ceil_ratio((*w_min)*(*h_max), s_unit*(*p_min))*s_unit;
	if (s > ss)  s = ss;
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_set_packet_size(camwire_handle_get_camera(c_handle),
					   video_mode,
					   s));
	ERROR_IF_DC1394_FAIL(
	    dc1394_format7_get_packets_per_frame(
		camwire_handle_get_camera(c_handle),
		video_mode,
		(uint32_t *)&pp));
	if (pp == 0)  pp = ceil_ratio((*w_min)*(*h_max), s);
	p = ceil_ratio((*w_min)*(*h_max), s);
	if (p != pp)
	    errorexit(c_handle, __func__,
		      "Camera's packet size calculation is suspect.");
	while (pp > *p_min)
	{
	    W = ceil_ratio(s_unit*pp, n_unit);
	    p = ceil_ratio(n_unit*W, s_unit);  /* Candidate p_min.*/
	    while (p != pp)  /* Find smallest integer solution. */
	    {
		pp = p;
		W = ceil_ratio(s_unit*pp, n_unit);
		p = ceil_ratio(n_unit*W, s_unit);
	    }
	    *w_min = (*w_unit)*W;
	    *p_min = p;
	    ERROR_IF_DC1394_FAIL(
		dc1394_format7_set_image_size(
		    camwire_handle_get_camera(c_handle),
		    video_mode,
		    *w_min, *h_max));
	    ERROR_IF_DC1394_FAIL(
		dc1394_format7_get_packet_parameters(camwire_handle_get_camera(c_handle),
						     video_mode,
						     (uint32_t *)&su, (uint32_t *)&ss));
	    if (su != s_unit)
		errorexit(c_handle, __func__, "Unit packet size has changed.");
	    if (ss > s_max)
		errorexit(c_handle, __func__, "Internal error: Bad s_max estimate.");
	    s = ceil_ratio((*w_min)*(*h_max), s_unit*(*p_min))*s_unit;
	    if (s > ss)  s = ss;
	    ERROR_IF_DC1394_FAIL(
		dc1394_format7_set_packet_size(camwire_handle_get_camera(c_handle),
					       video_mode,
					       s));
	    ERROR_IF_DC1394_FAIL(
		dc1394_format7_get_packets_per_frame(camwire_handle_get_camera(c_handle),
						     video_mode,
						     (uint32_t *)&pp));
	    if (pp == 0)  pp = ceil_ratio((*w_min)*(*h_max), s);
	    p = ceil_ratio((*w_min)*(*h_max), s);
	    if (p != pp)
		errorexit(c_handle, __func__,
			  "Camera's packet size calculation is suspect.");
	} /* while pp > *p_min */
    } /* if format */
/*     fprintf(stderr, "****p_min = %d, *w_min = %d\n", *p_min, *w_min);  // Debugging. */
} /* get_cam_limits() */

/*
  ----------------------------------------------------------------------
  Total transmission time measurement averaged over M frames, as a
  function of the number of packets p and the image height h and width
  w.  It includes the trigger setup time which cannot be measured
  directly.  The shutter (integration) time is varied randomly to avoid
  unintended synchronization with the bus cycle.
*/
static double T_total(const Camwire_handle c_handle,
		      const int p,
		      const int h,
		      const int w,
		      const int M,
		      const int ext_trig_OK,
		      const int serial_dev)
{
    dc1394video_mode_t video_mode;
    double framerate;
    /* double shutter; */
    double f_dummy;
    double T, trig_time, timestamp;
    double transmit_time_in_ts_calc;
    int new_width, new_height;
    int m;
/*     double pause_time; */
/*     struct timespec nap; */
    
    fprintf(stderr,
	    "Transmit time at p = %4d, h = %4d, w = %4d over %4d frames:",
	    p, h, w, M);
/*     fprintf(stderr, "\n");  // Debugging. */

    video_mode = convert_format_mode2dc1394video_mode(config.format, config.mode);

    /* Set camera to the wanted p and h, w: */
    if (variable_image_size(video_mode))  /* Format 7.*/
    {
	camwire_set_frame_offset(c_handle, 0, 0);
	camwire_set_frame_size(c_handle, w, h);
	camwire_get_frame_size(c_handle, &new_width, &new_height);
	if (new_width != w || new_height != h)
	    errorexit(c_handle, __func__, "Unexpected width or height change.");
    }
    if (variable_image_size(video_mode))  /* Format 7.*/
	framerate = numpackets2framerate(p);
    else
	framerate = 7200.0/p;
    camwire_set_framerate(c_handle, framerate);  /* Set the p.*/
    camwire_get_framerate(c_handle, &f_dummy);
    if (fabs(f_dummy - framerate) > 1e-08)
	errorexit(c_handle, __func__, "Unexpected framerate change.");
    
    /* Average over M frames with random shutter: */
    T = 0.0;
    for (m=0; m<M; ++m)
    {
/* 	/\* Some cameras don't want to be rushed: *\/ */
/* 	pause_time = (1.0 + uniform())/numpackets2framerate(p); */
/* 	nap.tv_sec = pause_time; */
/* 	nap.tv_nsec = (pause_time - nap.tv_sec)*1e9; */
/* 	nanosleep(&nap, 0); */

	/* Trigger: */
	/* shutter = */
	set_random_shutter(c_handle);
	if (ext_trig_OK)  trig_time = trigger(c_handle, serial_dev);
	else              trig_time = soft_trigger(c_handle);
/* 	fprintf(stderr, ".");  // Debugging. */

	/* Capture: */
	timestamp = get_image_timestamp(c_handle);
	transmit_time_in_ts_calc = assumed_transmit_time(c_handle, p, h);
	T += timestamp + transmit_time_in_ts_calc - trig_time;
	/* Both timestamp and trig_time contain the known shutter time
	   and config.trig_setup_time which therefore cancel in the
	   subtraction. */
    }
    T /= M;

    fprintf(stderr, "  %lg ms\n", T*1e03);
    return T;
} /* T_total() */

/*
  ----------------------------------------------------------------------
  Returns the random shutter set on the camera:
*/
static double set_random_shutter(const Camwire_handle c_handle)
{
    double believable, sh_range, shutter;

    /* Calculate range of at least 3 random shutters to span bus period.
       Having exposure_quantum a multiple of bus period would be
       unfortunate: */
    believable = 10*bus_period(config.bus_speed);
    if (config.exposure_quantum < believable && config.exposure_offset < believable)
    {  /* Normal-looking shutter.*/
	sh_range = max(3*config.exposure_quantum, bus_period(config.bus_speed));
	shutter = sh_range*uniform() + config.exposure_offset;
    }
    else
    {  /* We have an unusual shutter register.*/
	shutter = config.exposure_offset;
    }
    camwire_set_shutter(c_handle, shutter);
    camwire_get_shutter(c_handle, &shutter);
    return shutter;
} /* set_random_shutter() */

/*
  ----------------------------------------------------------------------
  Waits for an image to arrive and returns its timestamp as a double:
*/
static double get_image_timestamp(const Camwire_handle c_handle)
{
    void *buffer;
    int buffer_lag, num_buffers;
    struct timespec ts;

    camwire_point_next_frame(c_handle, &buffer, &buffer_lag);  /* Blocks.*/
    camwire_get_timestamp(c_handle, &ts);
    camwire_unpoint_frame(c_handle);
    if (!single_shot_capable)
    {  /* Camera has no single-shot mode.*/
	camwire_set_run_stop(c_handle, 0);
	complete_acquisition(c_handle);
	camwire_get_num_framebuffers(c_handle, &num_buffers);
	camwire_flush_framebuffers(c_handle, num_buffers, 0, &buffer_lag);
    }
    if (buffer_lag > 0)
	errorexit(c_handle, __func__, "Internal error: More than one frame pending.");
    return ts.tv_sec + ts.tv_nsec*1e-09;
} /* get_image_timestamp() */

/*
  ----------------------------------------------------------------------
  Help to undo Camwire's timestamp calculation using the camera's
  bus_speed, line_transfer_time and transmit_setup_time parameters, by
  recalculating the total transmission time assumed by Camwire.  This
  transmission time includes transmit_setup_time and trans_time, but not
  trig_setup_time or shutter.

  Camwire calculates the timestamp as (assuming juju stack):

      sensor_transfer_time = line_transfer_time*height;
      bus_transmit_time = bus_period(bus_speed)*num_packets;
      if (transmit_overlap)
	  trans_time = max(sensor_transfer_time, bus_transmit_time);
      else
	  trans_time = sensor_transfer_time + bus_transmit_time;
      trans_time -= bus_transmit_time;  // juju stack includes this already.
      timelag = trig_setup_time +
                shutter +
		transmit_setup_time +
		trans_time;
      timestamp = filltime - timelag;
*/
static double assumed_transmit_time(const Camwire_handle c_handle,
				    const int num_packets, const int height)
{
    Camwire_conf cam_config;
    double sensor_transfer_time, bus_transmit_time, trans_time;

    /* Get Camwire's own copy of the config parameters because we may
       have changed ours: */
    camwire_get_config(c_handle, &cam_config);
    sensor_transfer_time = cam_config.line_transfer_time*height;
    bus_transmit_time = bus_period(cam_config.bus_speed)*num_packets;
    if (cam_config.transmit_overlap)
	trans_time = max(sensor_transfer_time, bus_transmit_time);
    else
	trans_time = sensor_transfer_time + bus_transmit_time;
    trans_time -= bus_transmit_time;  /* juju stack includes this already.*/
    return cam_config.transmit_setup_time + trans_time;
} /* assumed_transmit_time() */

/*
  ----------------------------------------------------------------------
  Prints the given error message to stderr, cleans up the camera, and
  exits with a return status of 1.
*/
static void errorexit(const Camwire_handle c_handle, const char *funcname,
		      const char *msg)
{
    fprintf(stderr, "\n%s(): %s\n", funcname, msg);
    shutdown(c_handle);
    exit(1);
} /* errorexit() */

/*
  ----------------------------------------------------------------------
  Stops the camera and destroys the camera object and its bus.  Safe to
  call with a null argument.
*/
static void shutdown(const Camwire_handle c_handle)
{
    if (c_handle)
    {
	camwire_set_run_stop(c_handle, 0);
	complete_acquisition(c_handle);
    }
    camwire_destroy(c_handle);
    camwire_bus_destroy();
} /* shutdown() */

/*
  ----------------------------------------------------------------------
  Sleeps for more than one frame transmission period, including frame
  integration and sensor transfer time, so that any triggered frame
  acquision is completed on waking up again.
*/
static void complete_acquisition(const Camwire_handle c_handle)
{
    int height, width;
    int num_packets;
    double shutter, framerate;
    double sensor_transfer_time, bus_transmit_time, trans_time, acq_time;
    struct timespec nap;

    camwire_get_shutter(c_handle, &shutter);
    camwire_get_frame_size(c_handle, &height, &width);
    camwire_get_framerate(c_handle, &framerate);
    num_packets = framerate2numpackets(framerate);

    sensor_transfer_time = config.line_transfer_time*height;
    bus_transmit_time = bus_period(config.bus_speed)*num_packets;
    if (config.transmit_overlap)
	trans_time = max(sensor_transfer_time, bus_transmit_time);
    else
	trans_time = sensor_transfer_time + bus_transmit_time;
    acq_time = config.trig_setup_time + shutter + config.transmit_setup_time +
	trans_time;

    nap.tv_sec = acq_time;
    nap.tv_nsec = (acq_time - nap.tv_sec)*1e9;
    nanosleep(&nap, 0);
} /* complete_acquisition() */

/*
  -----------------------------------------------------------------------------
  Returns the frame rate corresponding to the given number of packets
  per frame.
*/
static double numpackets2framerate(const int num_packets)
{
    int actual;

    actual = num_packets;
    if (actual < 1)  actual = 1;
    if (actual > config.max_packets)  actual = config.max_packets;
    return(1.0/(actual*bus_period(config.bus_speed)));
} /* numpackets2framerate() */

/*
  -----------------------------------------------------------------------------
  Returns the number of video packets per frame corresponding to the
  given frame rate.
*/
static int framerate2numpackets(const double frame_rate)
{
    int num_packets;

    if (frame_rate <= 0)  return(config.max_packets);
    num_packets = (int) (1.0/(frame_rate*bus_period(config.bus_speed)) + 0.5);
    if (num_packets < 1)  num_packets = 1;
    if (num_packets > config.max_packets)  num_packets = config.max_packets;
    return(num_packets);
} /* framerate2numpackets() */

/*
  ----------------------------------------------------------------------
  Sends a rising edge on the DTR line of the serial port and returns its
  time as a double:
*/
static double trigger(const Camwire_handle c_handle, const int serial_dev)
{
    struct timespec downtime;
    struct timeval time1, time2;
    double t1, t2;

    /* First cause a small negative pulse: */
    downtime.tv_sec = 0;
    downtime.tv_nsec = 1000000;  /* 1 ms.*/
    setDTR(serial_dev, 0);
    nanosleep(&downtime, 0);

    /* Then produce the timed rising edge: */
    gettimeofday(&time1, 0);
    setDTR(serial_dev, 1);
    gettimeofday(&time2, 0);
    t1 = time1.tv_sec + time1.tv_usec*1e-06;
    t2 = time2.tv_sec + time2.tv_usec*1e-06;
    if (fabs(t2 - t1) > bus_period(config.bus_speed))
	errorexit(c_handle, __func__, "Tolerance on trigger time is too big.");

    /* The trigger remains normally high:*/
    return (t1 + t2)/2;
} /* trigger() */

/*
  ----------------------------------------------------------------------
  Sends a rising edge on the DTR line of the serial port and returns its
  time as a double:
*/
static double soft_trigger(const Camwire_handle c_handle)
{
    struct timespec downtime;
    struct timeval time1, time2;
    double t1, t2;
    double fiddle;

    /* First cause a small pause (to be compatible with trigger()): */
    downtime.tv_sec = 0;
    downtime.tv_nsec = 10000000;  /* 10 ms.*/
    nanosleep(&downtime, 0);

    /* Then trigger a single frame: */
    gettimeofday(&time1, 0);
    camwire_set_run_stop(c_handle, 1);  /* Assumes we are in single-shot mode.*/
    gettimeofday(&time2, 0);
    t1 = time1.tv_sec + time1.tv_usec*1e-06;
    t2 = time2.tv_sec + time2.tv_usec*1e-06;
    if (fabs(t2 - t1) > 20*bus_period(config.bus_speed))
    {
	errorexit(c_handle, __func__, "Tolerance on trigger time is too big.");
    }

    /* The camera captures a single frame and stops.  The returned
       trig_time is a guess at best.*/
    fiddle = 0.35;  /* Where we think trigger happened.*/
    return fiddle*t1 + (1 - fiddle)*t2;
} /* soft_trigger() */

/*
  ----------------------------------------------------------------------
Sets the state of the DTR line of the given file descriptor.
*/
static void setDTR(const int serial_dev, const int state)
{
    int status;

    ioctl(serial_dev, TIOCMGET, &status);
    if (state)  status |= TIOCM_DTR;
    else        status &= ~TIOCM_DTR;
    ioctl(serial_dev, TIOCMSET, &status);
} /* setDTR() */

/*
  ----------------------------------------------------------------------
  Calculate bus period from bus speed:
*/

static inline double bus_period(const int speed)
{ return 0.05/speed; }

/*
  ----------------------------------------------------------------------
  Numerical ceiling (rounding up) function for integer ratios:
*/

static inline int ceil_ratio(const int num, const int denom)
{ return (num + denom - 1)/denom; }

/*
  ----------------------------------------------------------------------
*/
static inline double max(const double a, const double b)
{ return (a > b ? a : b); }

/*
  ----------------------------------------------------------------------
*/
static inline double min(const double a, const double b)
{ return (a < b ? a : b); }

/*
  ----------------------------------------------------------------------
*/
static inline int xor(const int a, const int b)
{ return ((a!=0) != (b!=0)); }

/*
  ----------------------------------------------------------------------
  Accumulation for linear regression:
*/
static void acc(const double x, const double y, Linreg *rg)
{
    rg->sum_x  += x;
    rg->sum_y  += y;
    rg->sum_x2 += x*x;
    rg->sum_xy += x*y;
    rg->N++;
} /* acc() */

/*
  ----------------------------------------------------------------------
  Uniformly distributed random number in [0,1):
*/
static inline double uniform()
{
    return (double)rand()/((double)RAND_MAX + 1.0);
} /* uniform() */

/*
  ----------------------------------------------------------------------
  Choose a reasonable number of frames M (minimum 3) per measurement so
  that measurement does not take more than the given duration in
  seconds.
*/
static int choose_M(const int p)
{
    int max_p;
    int m;

    max_p = config.max_packets;
    m = (int)
	(measurement_duration/(sqrt(max_p*p)*bus_period(config.bus_speed)));
    if (m < 3)  return 3;  /* Do at least 3 measurements.*/
    return m;
} /* choose_M() */



/* FUNCTIONS COPIED FROM camwire.c: */
/* ******************************** */

/*
  -----------------------------------------------------------------------------
  Returns the dc1394 video_mode corresponding to the given numeric
  format and mode.  */
inline static dc1394video_mode_t convert_format_mode2dc1394video_mode(const int format,
								      const int mode)
{
    /* To translate to or from dc1394 mode enums: */
    static const int mode_dc1394_offset[] =
	{DC1394_VIDEO_MODE_160x120_YUV444,   /* Format 0.*/
	 DC1394_VIDEO_MODE_800x600_YUV422,   /* Format 1.*/
	 DC1394_VIDEO_MODE_1280x960_YUV422,  /* Format 2.*/
	 0, 0, 0,       /* Reserved formats.*/
	 DC1394_VIDEO_MODE_EXIF,             /* Format 6.*/
	 DC1394_VIDEO_MODE_FORMAT7_0};       /* Format 7.*/

    return mode_dc1394_offset[format] + mode;
}

/*
  -----------------------------------------------------------------------------
  Returns 1 (true) if the IEEE 1394 image format is a fixed image size,
  or 0 (false) otherwise.
*/
/*
inline static int fixed_image_size(const dc1394video_mode_t video_mode)
{
    return (dc1394_is_video_mode_scalable(video_mode) == DC1394_FALSE &&
	    dc1394_is_video_mode_still_image(video_mode) == DC1394_FALSE);
} // fixed_image_size() */

/*
  -----------------------------------------------------------------------------
  Returns 1 (true) if the IEEE 1394 image format is a variable image
  size, or 0 (false) otherwise.
*/
inline static int variable_image_size(const dc1394video_mode_t video_mode)
{
    return (dc1394_is_video_mode_scalable(video_mode) == DC1394_TRUE &&
	    dc1394_is_video_mode_still_image(video_mode) == DC1394_FALSE);
} /* variable_image_size() */
